@@ -1498,9 +1498,10 @@ class ParseIni(object):
 
         env_list = []
         envlist_explicit = False
-        if (from_option and "ALL" in from_option) or (
-            not from_option and from_environ and "ALL" in from_environ.split(",")
-        ):
+        if (
+            (from_option and "ALL" in from_option)
+            or (not from_option and from_environ and "ALL" in from_environ.split(","))
+        ) and PARALLEL_ENV_VAR_KEY_PRIVATE not in os.environ:
             all_envs = self._getallenvs(reader)
         else:
             candidates = (
@@ -1750,7 +1751,7 @@ class SectionReader:
     def getargv_install_command(self, name, default="", replace=True):
         s = self.getstring(name, default, replace=False)
         if not s:
-            # This occurs when factors are used, and a testenv doesnt have
+            # This occurs when factors are used, and a testenv doesn't have
             # a factorised value for install_command, most commonly occurring
             # if setting platform is also used.
             # An empty value causes error install_command must contain '{packages}'.
